#include <stdlib.h>
#include <stdio.h>
#include <GL/glut.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <iostream.h>
#include <fstream.h>

struct Space
{
  float x1, y1, x2, y2;
  Space *next;
};

struct Subsect
{
  char name[5];
  Space *spaces;
};

struct Wing 
{
  char name[5];
  Subsect subsection[10];
};

Wing Wallspace[4];

bool screengrabs = 0;

double min(double a, double b)
{
  return (a < b) ? a : b;
}

double max(double a, double b)
{
  return (a > b) ? a : b;
}

int pic = 0;
const float pi = 3.14159265358972;
float angle = 90.0;
float eyez = 3.0, eyex = 119.0;
float centerx = 0.0, centerz = 0.0;
int mainWin, mapWin;
int numofexhibits = 4;
int exhibit[5][2] = {{10, -5}, {7, 5}, {0, -5}, {-7, 5}, {-10, -5}};

/* Light Properties */
GLfloat light_ambient[] = {.75, .75, .75, 1};
GLfloat light_diffuse[] = {.35, .35, .35, 1};
GLfloat light_specular[] = {1.0, 1.0, 1.0, 1.0};
GLfloat light_position[] = {0, 25, 0, 0.2};
GLfloat map_light_position[] = {-1.0, 0.0, 0.0, 0.0};

char *number(int num)                             //changes int into array of chars
{
  int numdigits = int(floor(log(num) / log(10)));
  char *str = new char[4];
  str[3] = 0;
  for (int i=2; i>=0; i--)
  {
    if (i > numdigits) str[2-i] = '0';
    else
    {
      int digit = int(num/pow(10,i));
      str[2-i] = (char)(48+digit);
      num -= int(digit*pow(10,i));
    }
  }
  return str;
}


int screenShot()
{
  unsigned char *pixels;                          // store the image data here
  FILE * shot;                                    // use to write files
  int screenStats[4];                             // width/height of screen

  glGetIntegerv(GL_VIEWPORT, screenStats);        // width/height of window

                                                  // make array to hold the pixel data:
                                                  // width*height*bytesperpixel
  pixels = new unsigned char[screenStats[2]*screenStats[3]*3];
                                                  // read in the pixel data
  glReadPixels(0, 0, screenStats[2], screenStats[3], GL_BGR, 
                                   GL_UNSIGNED_BYTE, pixels);

  char *filename = new char[12];                  // where to save the input file: img/***.tga
  const char *temp = number(pic/2);
  filename[0]='i';
  filename[1]='m';
  filename[2]='g';
  filename[3]='/';
  for(int x=0; x<3; x++)
    filename[x+4]=temp[x];
  filename[7]='.';
  filename[8]='t';
  filename[9]='g';
  filename[10]='a';
  filename[11]=0;

                                                  // open file for writing-if fail, return 1
  if((shot=fopen(filename, "wb"))==NULL) return 1;

                                                  // tga header--must begin every .tga
  unsigned char TGAheader[12]={0,0,2,0,0,0,0,0,0,0,0,0};
  unsigned char header[6]={((int)(screenStats[2]%256)),
                           ((int)(screenStats[2]/256)),
                           ((int)(screenStats[3]%256)),
                           ((int)(screenStats[3]/256)),24,0};

                                                  // write out the TGA header
  fwrite(TGAheader, sizeof(unsigned char), 12, shot);
  fwrite(header, sizeof(unsigned char), 6, shot); // write out the header
                                                  // write the pixels
  fwrite(pixels, sizeof(unsigned char), screenStats[2]*screenStats[3]*3, shot);

  fclose(shot);                                   // close the file
  delete [] pixels;                               // free the memory

  return 0;
}

static GLuint textureIDs[9];

GLuint readTexture(const char* filename)          // reads in a texture from filename
{
  ifstream infile(filename);
  if(infile.fail()) exit(0);
  do 
  {
    infile.ignore(200, '\n');
  } 
  while(infile.peek() == '#');
  int width, height, depth;
  infile >> width >> height >> depth;
  GLubyte* texMap = new GLubyte[width * height * 3];
  int temp;
  for(int i = 0; i < height * width * 3; i++)
  {
    infile >> temp;
    texMap[i] = char(temp);
  }
  infile.close();
  GLuint texID;
  glGenTextures(1, &texID);
  glBindTexture(GL_TEXTURE_2D, texID);
  gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, width, height, GL_RGB, GL_UNSIGNED_BYTE, texMap);
  delete[] texMap;                                // otherwise it'll be in memory twice
  return texID;
}

void textureInit()                                // input all textures used
{
  glShadeModel(GL_SMOOTH);
  glEnable(GL_DEPTH_TEST);
  textureIDs[0] = readTexture("door1.ppm");
  textureIDs[1] = readTexture("floor2.ppm");
  textureIDs[2] = readTexture("walls1.ppm");
  textureIDs[3] = readTexture("ceiling1.ppm");
  textureIDs[4] = readTexture("exhibitflappers.ppm");
  textureIDs[5] = readTexture("exhibitdepression.ppm");
  textureIDs[6] = readTexture("exhibithindenburg.ppm");
  textureIDs[7] = readTexture("exhibitmccarthy.ppm");
  textureIDs[8] = readTexture("exhibitapollo.ppm");
  glPixelStorei(GL_UNPACK_ALIGNMENT, GL_UNSIGNED_BYTE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
}

void drawTexture(int texNum, float x1, float y1, float z1, float x2, float y2, float z2)
{
  glEnable(GL_TEXTURE_2D);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glBindTexture(GL_TEXTURE_2D, textureIDs[texNum]);
  if(x1 == x2)
  {
    glBegin(GL_QUADS);
      glNormal3f( 1.0, 0.0, 0.0 );
      glTexCoord2f(0.0, 0.0); glVertex3f(x1, y1, z1);
      glTexCoord2f(1.0, 0.0); glVertex3f(x1, y1, z2);
      glTexCoord2f(1.0, 1.0); glVertex3f(x1, y2, z2);
      glTexCoord2f(0.0, 1.0); glVertex3f(x1, y2, z1);
    glEnd();
  }
  else if(y1 == y2)
  {
    glBegin(GL_QUADS);
      glNormal3f( 0.0, -1.0, 0.0 );
      glTexCoord2f(0.0, 0.0); glVertex3f(x1, y1, z1);
      glTexCoord2f(1.0, 0.0); glVertex3f(x2, y1, z1);
      glTexCoord2f(1.0, 1.0); glVertex3f(x2, y1, z2);
      glTexCoord2f(0.0, 1.0); glVertex3f(x1, y1, z2);
    glEnd();
  }
  else if(z1 == z2)
  {
    glBegin(GL_QUADS);
      glNormal3f( 0.0, 0.0, 1.0 );
      glTexCoord2f(0.0, 0.0); glVertex3f(x1, y1, z1);
      glTexCoord2f(1.0, 0.0); glVertex3f(x2, y1, z1);
      glTexCoord2f(1.0, 1.0); glVertex3f(x2, y2, z1);
      glTexCoord2f(0.0, 1.0); glVertex3f(x1, y2, z1);
    glEnd();
  }
  glFlush();
  glDisable(GL_TEXTURE_2D);
}

void displayFont(char* str)                       // write something on museum walls
{
  int len = strlen(str);
  for(int i = 0; i < len; i++)
    glutStrokeCharacter(GLUT_STROKE_ROMAN, int(str[i]));
}

                                                  // tile wall texture on rectangle given
void walltextures(float x1, float z1, float x2, float z2)
{
  int tempinc = 0;
  double inc = 0;
  if(z1 == z2) 
  {
    tempinc = ((x2-x1)/5) + 1;	
    inc = (x2-x1)/tempinc;
  }
  else 
  {
    tempinc = ((z2-z1)/5) + 1;	
    inc = (z2-z1)/tempinc;
  }
  if (inc < 0) inc = inc * (-1.0);
	
  for(int w=0; w<tempinc; w++) 
  {
    if(z1 == z2)
    {
      drawTexture(2, x1 + (w*inc), .5, z1, x1 + (w+1)*inc, 5.5, z1);
      drawTexture(2, x1 + (w*inc), 5.5, z1, x1 + (w+1)*inc, 10.5, z1);
      drawTexture(2, x1 + (w*inc), 10.5, z1, x1 + (w+1)*inc, 15.5, z1);
    }
    else
    {
      drawTexture(2, x1, .5, z1 + (w*inc), x1, 5.5, z1 + (1+w)*inc);
      drawTexture(2, x1, 5.5, z1 + (w*inc), x1, 10.5, z1 + (1+w)*inc);
      drawTexture(2, x1, 10.5, z1 + (w*inc), x1, 15.5, z1 + (1+w)*inc);
    }
  }
}

void drawCeiling()                                // tile ceiling texture on ceiling
{
  glNormal3f( 0.0, -1.0, 0.0 );
  for(int i = 0; i < 17; i++)
    for(int j = 0; j < 92; j++)
      drawTexture(3, j*3, 12, i*10, (j+1)*3, 12, (i+1)*10);
}

void drawFloors()                                 // tile floor texture on floor
{
  glNormal3f( 0.0, 1.0, 0.0 );
  for(int i = 0; i < 17; i++)
    for(int j = 0; j < 55; j++)
      drawTexture(1, j*5, .51, i*10, (j+1)*5, .51, (i+1)*10);
}

void drawDoors()                                  // draw the doors in the museum
{
  drawTexture(0, 117.5, .5, 1.15, 120.5, 7, 1.15);
  drawTexture(0, 88.15, .5, 22, 88.15, 7, 25);
  drawTexture(0, 27, .5, 47.15, 30, 7, 47.15);
}			

void drawHalls(void)                              // draw the walls in the museum
{
  float x1, x2, z1, z2;
  ifstream infile("walls.txt");                   // read in file that gives wall coordinates
  if(infile.fail())  exit(0);
  while (infile >> x1 >> z1 >> x2 >> z2)          // reads in coordinates of 2 opposite corners
  {
    walltextures(x1, z1, x2, z1);                 // puts texture on all four faces of walls
    walltextures(x1, z2, x2, z2);
    walltextures(x1, z1, x1, z2);
    walltextures(x2, z1, x2, z2);
  }
  infile.close();

  ifstream innfile("openings.txt");               // fill in areas above openings between rooms
  if(innfile.fail())  exit(0);
  while(innfile >> x1 >> z1 >> x2 >> z2)
  {
    drawTexture(2, x1, 8, z1, x1, 13, z2);
    drawTexture(2, x2, 8, z1, x2, 13, z2);
    drawTexture(2, x1, 8, z1, x2, 13, z1);
    drawTexture(2, x1, 8, z2, x2, 13, z2);
    drawTexture(2, x1, 8, z1, x2, 8, z2);
  }
}

void display(void)
{
  centerx = eyex + 5.0 * cos(angle * pi / 180.0);
  centerz = eyez + 5.0 * sin(angle * pi / 180.0);

  gluPerspective( 60.0, 1.0 , 0.1,  170.0);
  glMatrixMode(GL_MODELVIEW);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();
  glClearColor(0.4375, 0.78125, 0.9414, 0.0);     // sky blue

  glPushMatrix();

  /* Light */
  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);

  /* Enable Various Components */
  glEnable(GL_COLOR_MATERIAL);  
  glEnable(GL_LIGHT0);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);
  GLfloat shine[] = {0.2, 0.2, 0.2};
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, shine);
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

  gluLookAt(eyex, 5.0, eyez, centerx, 5.0, centerz, 0.0, 1.0, 0.0);

  drawHalls();
  drawFloors();
  drawCeiling();
  drawDoors();

  glPopMatrix();
  glFlush();
  glutSwapBuffers();
  glutPostRedisplay();
}

void myReshape(int w, int h)                      // reshape function
{
  glViewport(0, 0, w, h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();   
  gluPerspective( 40.0, (GLfloat) h / (GLfloat) w , 0.1,  170.0);
  glMatrixMode(GL_MODELVIEW);
}

void keyboard(int key, int x, int y)              // keyboard is used for movement
{
  if(key == GLUT_KEY_LEFT) angle -= 3.0;
  else if(key == GLUT_KEY_RIGHT) angle += 3.0;
  else if(key == GLUT_KEY_UP) 
  {
    eyex = eyex + 0.5 * cos(angle * pi / 180.0);
    eyez = eyez + 0.5 * sin(angle * pi / 180.0);
  }
  else if(key == GLUT_KEY_DOWN) 
  {
    eyex = eyex - 0.5 * cos(angle * pi / 180.0);
    eyez = eyez - 0.5 * sin(angle * pi / 180.0);
  }
  glutPostRedisplay();

  if(pic < 200 && screengrabs)                    // take screengrabs
  {
	if(pic%2==0) screenShot();
	pic++;
  }
}

void mouse(int button, int state, int x, int y)   // when user clicks on an exhibit
{
  char* exhibitCall = new char[100];
  if (state == GLUT_DOWN)
  {
    for(int i = 0; i <= numofexhibits; i++)
    {
      if((exhibit[i][0] - centerx) * (exhibit[i][0] - centerx) + 
         (exhibit[i][1] - centerz) * (exhibit[i][1] - centerz) < 3)
      {
        cout << "Opening exhibit #" << i + 1 << endl;
        sprintf(exhibitCall, "mozilla http://www.tjhsst.edu/~lpham/supercomp/imagine/exhibit%i.htm", i);
        cout << exhibitCall << endl;
        system(exhibitCall);
        cout << "Exhibit #" << i + 1<< " closed\n";
      }
    }
  }
}

void fogInit()                                    // create fog to show distance
{
  glEnable(GL_FOG);
  GLfloat fogcol[] = {.4, .4, .4, 1};
  glFogi(GL_FOG_MODE, GL_EXP2);
  glFogfv(GL_FOG_COLOR, fogcol);
  glFogf(GL_FOG_DENSITY, 0.01);
  glFogf(GL_FOG_START, .1);
  glFogf(GL_FOG_END, 150);
}

int main(int argc, char **argv)
{
  int width, height;
  if(screengrabs)                                 // if taking screengrabs, make small window
  {
    width = 256;
    height = 256;
  }
  else
  {
    width = 512;
    height = 512;
  }
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
  glutInitWindowPosition(50, 50);
  glutInitWindowSize(width, height);
  mainWin = glutCreateWindow("VirtualMuseum");
  textureInit();
  fogInit();
  glutMouseFunc(mouse);
  glutSpecialFunc(keyboard);
  glutReshapeFunc(myReshape);
  glutDisplayFunc(display);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
  glutMainLoop();
  return 0;
}
